/******************************************************************
Copyright 2000 by Object Craft P/L, Melbourne, Australia.

                        All Rights Reserved

Permission to use, copy, modify, and distribute this software and its
documentation for any purpose and without fee is hereby granted,
provided that the above copyright notice appear in all copies and that
both that copyright notice and this permission notice appear in
supporting documentation, and that the name of Object Craft
is not be used in advertising or publicity pertaining to
distribution of the software without specific, written prior
permission.

OBJECT CRAFT DISCLAIMS ALL WARRANTIES WITH REGARD TO THIS SOFTWARE,
INCLUDING ALL IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS, IN NO
EVENT SHALL OBJECT CRAFT BE LIABLE FOR ANY SPECIAL, INDIRECT OR
CONSEQUENTIAL DAMAGES OR ANY DAMAGES WHATSOEVER RESULTING FROM LOSS OF
USE, DATA OR PROFITS, WHETHER IN AN ACTION OF CONTRACT, NEGLIGENCE OR
OTHER TORTIOUS ACTION, ARISING OUT OF OR IN CONNECTION WITH THE USE OR
PERFORMANCE OF THIS SOFTWARE.

******************************************************************/

#include "mssqldb.h"

static PyObject *ErrorObject;

static char mssqldb_DataBuf__doc__[] =
"DataBuf(type [, len]) -> databuf";

static PyObject *mssqldb_DataBuf(PyObject *self, PyObject *args)
{
    int vartype, varlen = 0;

    if (!PyArg_ParseTuple(args, "i|i", &vartype, &varlen))
	return NULL;

    return DataBuf__init__(vartype, varlen);
}

static char mssqldb_dbconvert__doc__[] =
"dbconvert(obj, databuf) -> status";

static PyObject *mssqldb_dbconvert(PyObject *self, PyObject *args)
{
    PyObject *from;
    DataBufObj *to;

    if (!PyArg_ParseTuple(args, "OO!", &from, &DataBufType, &to))
	return NULL;
    return DataBuf_convert(NULL, from, to);
}

static PyObject *err_handler_cb;

static int err_handler(DBPROCESS *db_dbproc,
		       int severity, int dberr, int oserr,
		       const char *dberrstr, const char *oserrstr)
{
    PyObject *dbproc, *args, *result;

    if (db_dbproc == NULL)
	dbproc = Py_None;
    else {
	dbproc = dbproc_find_object(db_dbproc);
	if (dbproc == NULL)
	    return INT_CANCEL;
    }
    if (oserr == DBNOERR)
	oserrstr = "";
    args = Py_BuildValue("(Oiiiss)",
			 dbproc, severity, dberr, oserr, dberrstr, oserrstr);
    if (args == NULL)
	return INT_CANCEL;
    result = PyEval_CallObject(err_handler_cb, args);
    Py_DECREF(args);
    if (result != NULL) {
	RETCODE status = INT_CANCEL;

	if (PyInt_Check(result))
	    status = PyInt_AsLong(result);
	Py_DECREF(result);
	return status;
    }

    return INT_CANCEL;
}

static PyObject *msg_handler_cb;

static int msg_handler(DBPROCESS *db_dbproc,
		       DBINT msgno, int msgstate, int severity,
		       const char *msgtext, const char *srvname, const char *procname,
		       DBUSMALLINT line)
{
    PyObject *dbproc, *args, *result;

    if (db_dbproc == NULL)
	dbproc = Py_None;
    else {
	dbproc = dbproc_find_object(db_dbproc);
	if (dbproc == NULL)
	    dbproc = Py_None;
    }
    args = Py_BuildValue("(Oiiisssi)",
			 dbproc, msgno, msgstate, severity,
			 msgtext, srvname, procname, line);
    if (args == NULL)
	return 0;
    result = PyEval_CallObject(msg_handler_cb, args);
    Py_DECREF(args);
    if (result != NULL) {
	RETCODE status = 0;

	if (PyInt_Check(result))
	    status = PyInt_AsLong(result);
	Py_DECREF(result);
	return status;
    }

    return 0;
}

static char mssqldb_dberrhandle__doc__[] =
"dberrhandle(callable) -> callable";

static PyObject *mssqldb_dberrhandle(PyObject *self, PyObject *args)
{
    static int handler_set;
    PyObject *func;
    PyObject *old_func;

    if (!PyArg_ParseTuple(args, "O", &func))
	return NULL;

    old_func = err_handler_cb;
    if (func == Py_None)
	err_handler_cb = NULL;
    else {
	if (!PyCallable_Check(func)) {
	    PyErr_SetString(PyExc_TypeError, "parameter must be callable");
	    return NULL;
	}
	err_handler_cb = func;
	Py_XINCREF(err_handler_cb);
    }

    if (!handler_set) {
	handler_set = 1;
	PyErr_Clear();
#ifdef _WIN32
	dberrhandle(err_handler);
#else
	dberrhandle((EHANDLEFUNC)err_handler);
#endif
	if (PyErr_Occurred())
	    return NULL;
    }

    if (old_func != NULL)
	return old_func;
    Py_INCREF(Py_None);
    return Py_None;
}

static char mssqldb_dbexit__doc__[] =
"dbexit() -> None";

static PyObject *mssqldb_dbexit(PyObject *self, PyObject *args)
{
    if (!PyArg_ParseTuple(args, ""))
	return NULL;

    PyErr_Clear();
    dbexit();
    if (PyErr_Occurred())
	return NULL;

    Py_INCREF(Py_None);
    return Py_None;
}

static char mssqldb_dbinit__doc__[] =
"dbinit() -> bool";

#ifdef _WIN32
static PyObject *mssqldb_dbinit(PyObject *self, PyObject *args)
{
    const char *version;

    if (!PyArg_ParseTuple(args, ""))
	return NULL;

    PyErr_Clear();
    version = dbinit();
    if (PyErr_Occurred())
	return NULL;

    if (version == NULL) {
	Py_INCREF(Py_None);
	return Py_None;
    }

    return PyString_FromString(version);
}
#else
static PyObject *mssqldb_dbinit(PyObject *self, PyObject *args)
{
    RETCODE status;

    if (!PyArg_ParseTuple(args, ""))
	return NULL;

    PyErr_Clear();
    status = dbinit();
    if (PyErr_Occurred())
	return NULL;

    return PyInt_FromLong(status);
}
#endif

static char mssqldb_dblogin__doc__[] =
"dblogin() -> LOGINREC";

static PyObject *mssqldb_dblogin(PyObject *self, PyObject *args)
{
    if (!PyArg_ParseTuple(args, ""))
	return NULL;

    return LOGINREC__init__();
}

static char mssqldb_dbmsghandle__doc__[] =
"dbmsghandle(callable) -> callable";

static PyObject *mssqldb_dbmsghandle(PyObject *self, PyObject *args)
{
    static int handler_set;
    PyObject *func;
    PyObject *old_func;

    if (!PyArg_ParseTuple(args, "O", &func))
	return NULL;

    old_func = msg_handler_cb;
    if (func == Py_None)
	msg_handler_cb = NULL;
    else {
	if (!PyCallable_Check(func)) {
	    PyErr_SetString(PyExc_TypeError, "parameter must be callable");
	    return NULL;
	}
	msg_handler_cb = func;
	Py_XINCREF(msg_handler_cb);
    }

    if (!handler_set) {
	handler_set = 1;
	PyErr_Clear();
#ifdef _WIN32
	dbmsghandle(msg_handler);
#else
	dbmsghandle((MHANDLEFUNC)msg_handler);
#endif
	if (PyErr_Occurred())
	    return NULL;
    }

    if (old_func != NULL)
	return old_func;
    Py_INCREF(Py_None);
    return Py_None;
}

static char mssqldb_dbprtype__doc__[] = 
"dbprtype(token) -> name";

static PyObject *mssqldb_dbprtype(DBPROCObj *self, PyObject *args)
{
    int token;
    const char *str;

    if (!PyArg_ParseTuple(args, "i", &token))
	return NULL;

    str = dbprtype(token);

    if (str == NULL) {
	Py_INCREF(Py_None);
	return Py_None;
    }
    return PyString_FromString(str);
}

static struct PyMethodDef mssqldb_methods[] = {
    { "DataBuf", (PyCFunction)mssqldb_DataBuf, METH_VARARGS, mssqldb_DataBuf__doc__ },
    { "datetime", (PyCFunction)DateTimeType_new, METH_VARARGS, datetime_new__doc__ },
    { "dbconvert", (PyCFunction)mssqldb_dbconvert, METH_VARARGS, mssqldb_dbconvert__doc__ },
    { "dberrhandle", (PyCFunction)mssqldb_dberrhandle, METH_VARARGS, mssqldb_dberrhandle__doc__ },
    { "dbexit", (PyCFunction)mssqldb_dbexit, METH_VARARGS, mssqldb_dbexit__doc__ },
    { "dbinit", (PyCFunction)mssqldb_dbinit, METH_VARARGS, mssqldb_dbinit__doc__ },
    { "dblogin", (PyCFunction)mssqldb_dblogin, METH_VARARGS, mssqldb_dblogin__doc__ },
    { "dbmsghandle", (PyCFunction)mssqldb_dbmsghandle, METH_VARARGS, mssqldb_dbmsghandle__doc__ },
    { "dbprtype", (PyCFunction)mssqldb_dbprtype, METH_VARARGS, mssqldb_dbprtype__doc__ },
    { "money", (PyCFunction)MoneyType_new, METH_VARARGS, money_new__doc__ },
    { "numeric", (PyCFunction)NumericType_new, METH_VARARGS, numeric_new__doc__ },
    { "pickle_datetime", (PyCFunction)pickle_datetime, METH_VARARGS, pickle_datetime__doc__ },
    { "pickle_money", (PyCFunction)pickle_money, METH_VARARGS, pickle_money__doc__ },
    { "pickle_numeric", (PyCFunction)pickle_numeric, METH_VARARGS, pickle_numeric__doc__ },
    { NULL, (PyCFunction)NULL, 0, NULL }
};

#define DBVAL(t, v) { t, #v, v }

typedef struct {
    int type;
    char *name;
    int value;
} value_desc;

static value_desc mssql_args[] = {
    DBVAL(TYPE_RETCODE, FAIL),
    DBVAL(TYPE_RETCODE, SUCCEED),
    DBVAL(TYPE_RETCODE, NO_MORE_RESULTS),

    DBVAL(TYPE_ANSWER, INT_EXIT),
    DBVAL(TYPE_ANSWER, INT_CONTINUE),
    DBVAL(TYPE_ANSWER, INT_CANCEL),
#ifdef INT_TIMEOUT
    DBVAL(TYPE_ANSWER, INT_TIMEOUT),
#endif

    DBVAL(TYPE_ROWRES, MORE_ROWS),
    DBVAL(TYPE_ROWRES, NO_MORE_ROWS),
    DBVAL(TYPE_ROWRES, REG_ROW),
    DBVAL(TYPE_ROWRES, BUF_FULL),

    DBVAL(TYPE_VALUE, SQLMAXNAME),

    DBVAL(TYPE_SERVTYPE, SQLVOID),
    DBVAL(TYPE_SERVTYPE, SQLTEXT),
    DBVAL(TYPE_SERVTYPE, SQLVARBINARY),
    DBVAL(TYPE_SERVTYPE, SQLINTN),
    DBVAL(TYPE_SERVTYPE, SQLVARCHAR),
    DBVAL(TYPE_SERVTYPE, SQLBINARY),
    DBVAL(TYPE_SERVTYPE, SQLIMAGE),
    DBVAL(TYPE_SERVTYPE, SQLCHAR),
    DBVAL(TYPE_SERVTYPE, SQLINT1),
    DBVAL(TYPE_SERVTYPE, SQLBIT),
    DBVAL(TYPE_SERVTYPE, SQLINT2),
    DBVAL(TYPE_SERVTYPE, SQLINT4),
    DBVAL(TYPE_SERVTYPE, SQLMONEY),
    DBVAL(TYPE_SERVTYPE, SQLDATETIME),
    DBVAL(TYPE_SERVTYPE, SQLFLT8),
    DBVAL(TYPE_SERVTYPE, SQLFLTN),
    DBVAL(TYPE_SERVTYPE, SQLMONEYN),
    DBVAL(TYPE_SERVTYPE, SQLDATETIMN),
    DBVAL(TYPE_SERVTYPE, SQLFLT4),
    DBVAL(TYPE_SERVTYPE, SQLMONEY4),
    DBVAL(TYPE_SERVTYPE, SQLDATETIM4),
    DBVAL(TYPE_SERVTYPE, SQLDECIMAL),
    DBVAL(TYPE_SERVTYPE, SQLNUMERIC),

#ifndef _WIN32
    DBVAL(TYPE_DATEBIT, DBDATE_YY),
    DBVAL(TYPE_DATEBIT, DBDATE_QQ),
    DBVAL(TYPE_DATEBIT, DBDATE_MM),
    DBVAL(TYPE_DATEBIT, DBDATE_DW),
    DBVAL(TYPE_DATEBIT, DBDATE_DD),
    DBVAL(TYPE_DATEBIT, DBDATE_DY),
    DBVAL(TYPE_DATEBIT, DBDATE_HH),
    DBVAL(TYPE_DATEBIT, DBDATE_MI),
    DBVAL(TYPE_DATEBIT, DBDATE_SS),
    DBVAL(TYPE_DATEBIT, DBDATE_MS),
    DBVAL(TYPE_DATEBIT, DBDATE_WK),
    DBVAL(TYPE_DATEBIT, DBDATE_CY),
    DBVAL(TYPE_DATEBIT, DBDATE_CW),
    DBVAL(TYPE_DATEBIT, DBDATE_CD),
#endif

    DBVAL(TYPE_RPCOPT, DBRPCRECOMPILE),
#ifdef DBNOTIFYALL
    DBVAL(TYPE_RPCOPT, DBNOTIFYALL),
#endif
#ifdef DBNOTIFYNEXT
    DBVAL(TYPE_RPCOPT, DBNOTIFYNEXT),
#endif

#ifdef DBRPCNORETURN
    DBVAL(TYPE_RPCPARAM, DBRPCNORETURN),
#endif
    DBVAL(TYPE_RPCPARAM, DBRPCRETURN),
#ifdef DB__NODEFAULT
    DBVAL(TYPE_RPCPARAM, DB__NODEFAULT),
#endif

    {0,0,0}
};

static int dict_add_int(PyObject *dict, char *key, int value)
{
    int err;
    PyObject *obj = PyInt_FromLong(value);
    if (obj == NULL)
	return 0;

    err = PyDict_SetItemString(dict, key, obj);
    Py_DECREF(obj);
    return !err;
}

static int dict_add_type(PyObject *dict, PyTypeObject *type)
{
    return PyDict_SetItemString(dict, type->tp_name, (PyObject*)type);
}

static char mssqldb_module_documentation[] = 
"";

void initmssqldb(void)
{
    PyObject *m, *d;
    value_desc *desc;

    LOGINRECType.ob_type = &PyType_Type;
    DBPROCType.ob_type = &PyType_Type;
    DataBufType.ob_type = &PyType_Type;
    NumericType.ob_type = &PyType_Type;
    DateTimeType.ob_type = &PyType_Type;
    MoneyType.ob_type = &PyType_Type;

    m = Py_InitModule4("mssqldb", mssqldb_methods,
		       mssqldb_module_documentation,
		       (PyObject*)NULL, PYTHON_API_VERSION);

    d = PyModule_GetDict(m);
    /* Add constants */
    for (desc = mssql_args; desc->name != NULL; desc++)
	if (!dict_add_int(d, desc->name, desc->value))
	    break;

    /* Add type objects */
    if (dict_add_type(d, &LOGINRECType)
	|| dict_add_type(d, &DBPROCType)
	|| dict_add_type(d, &DataBufType)
	|| dict_add_type(d, &DateTimeType)
	|| dict_add_type(d, &NumericType))
	;

    /* Register the numeric pickler */
    copy_reg_numeric(d);
    copy_reg_datetime(d);
    copy_reg_money(d);

    ErrorObject = PyString_FromString("mssqldb.error");
    PyDict_SetItemString(d, "error", ErrorObject);

    if (PyErr_Occurred())
	Py_FatalError("can't initialize module mssqldb");
}

